/****************************************************************************

Package:            de.mvcsys.epocutil
Class:              EFrame

Task:               Frame for EPOC applications

Author:             Jonathan Allin
Modified by:        Marcus von Cube

Credits:            Borrowed from "Professional Symbian Programming"
                    by Martin Tasker, Birmingham 2000

Updates:

Datum     Name  Vers.  Bemerkung
-----------------------------------------------------------------------------
25.04.99  JA    0.1    created
26.04.03  MvC   0.2    package de.mvcsys.epocutil, toolbar
21.11.03  MvC   0.55   setButtonLabel fixed (wrong index)

****************************************************************************/

package de.mvcsys.epocutil;

import de.mvcsys.awtutil.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.text.*;

/**
 *  <p>The EFrame class is intended to provide some of the functionality
 *  needed to make a Java application look and behave like a C++ EPOC
 *  application.</p>
 *
 *  <p>So far EFrame provides:</p>
 *  <ul>
 *    <li>Full screen</li>
 *    <li>Menu bar with Exit and About entries (these will need to be
 *        internationalised)</li>
 *    <li>An About dialog, with methods aboutDialog.setTitle(String) and
 *        aboutDialog.setText(String[])</li>
 *    <li>Window closing handler</li>
 *  </ul>
 *
 *  <p>A Java application can extend EFrame, rather than Frame,
 *  to use this additional functionality.</p>
 *
 *  <p>Additional 'E' widgets to develop include:</p>
 *  <ul>
 *     <li>a Dialog with a layout manager that conforms to EPOC guidelines</li>
 *     <li>a {@link EButton Button} which includes a short cut legend</li>
 *  </ul>
 *
 *  @author  Jonathan Allin. (C) Jonathan Allin 1999
 *  @version 0.1, 25/04/99
 */
public class EFrame extends Frame implements ActionListener
{
    public static final int TINY   = 0;
    public static final int NORMAL = 1;
    public static final int LARGE  = 2;

    public static int defaultSizeTag = TINY;
    public int sizeTag = defaultSizeTag;

    public static final Font MENU_FONT =
            new Font( "Dialog", Font.BOLD, 14 );

    public boolean isEpoc = false;

    public Panel contentPanel;
    public Toolbar toolbar;
    public AboutDialog aboutDialog;

    public MenuBar menuBar;
    public Menu fileMenu;
    public MenuItem closeItem;
    public Menu viewMenu;
    public MenuItem toolbarItem;
    public Menu toolsMenu;
    public MenuItem aboutItem;
    public MenuItem helpItem;
    public String helpDocument;

    private static final int BORDER_WIDTH = 5;
    private static final int TITLE_HEIGHT = 26;

    protected int frameX;
    protected int frameY;
    protected int frameWidth;
    protected int frameHeight;

    /**
     *  Default Constructor with standard title text
     */
    public EFrame()
    {
        this( "JavaApp" );
    }


    /**
     *  Contructor to set the window size to a reasonable EPOC maximum,
     *  hiding the frame boundaries.
     *
     *  A default menu bar is created.
     *
     *  @param title the (on EPOC invisible) window title
     */
    public EFrame( String title )
    {
        this( title, null );
    }


    /**
     *  Contructor to set the window size to a reasonable EPOC maximum,
     *  hiding the frame boundaries.
     *
     *  A default menu bar and toolbar is created.
     *
     *  @param title the window and toolbar title
     *  @param buttons array of button texts for the toolbar
     */
    public EFrame( String title, String[] buttons )
    {
        super( title );

        try {
            isEpoc = EpocUtil.isEpoc();
        }
        catch ( Exception e ) {}

        if ( isEpoc ) {
            Dimension screenSize = getToolkit().getScreenSize();
            sizeTag = screenSize.height < 240 ? TINY
                    : screenSize.height < 480 ? NORMAL
                                              : LARGE;
            frameX = -BORDER_WIDTH;
            frameY = -TITLE_HEIGHT;
            frameWidth  = screenSize.width + 2 * BORDER_WIDTH;
            frameHeight = screenSize.height  + TITLE_HEIGHT + BORDER_WIDTH;
        }
        else if ( EpocUtil.isDebug() ) {
            System.out.println( System.getProperty( "os.name" ) );
            System.out.println( DateFormat.getDateTimeInstance(
                                    DateFormat.FULL, DateFormat.FULL ).
                                    format( new Date() ) );
        }

        /*
         *  everything except the toolbar goes here
         */
        contentPanel = new Panel();
        contentPanel.setLayout( new BorderLayout() );
        super.add( contentPanel, BorderLayout.CENTER );

        if ( buttons != null ) {
            toolbar = new Toolbar( title, buttons );
            super.add( toolbar, BorderLayout.EAST );
        }

        /*
         *  creawte default menu
         */
        aboutDialog = new AboutDialog( this );

        menuBar = new MenuBar();

        fileMenu = new Menu( "File" );
        addUserItems( fileMenu );
        closeItem = addMenuItem( fileMenu, "Close", KeyEvent.VK_E );
        addMenu( fileMenu );

        viewMenu = new Menu( "View" );
        addUserItems( viewMenu );
        if ( toolbar != null ) {
            toolbarItem =
                    addMenuItem( viewMenu, "Show toolbar", KeyEvent.VK_T );
        }
        addMenu( viewMenu );

        toolsMenu = new Menu( "Tools" );
        addUserItems( toolsMenu );
        aboutItem = addMenuItem( toolsMenu, "About\u2026", KeyEvent.VK_A );
        helpItem = addMenuItem( toolsMenu, "Help", KeyEvent.VK_H );

        addMenu( toolsMenu );

        setMenuBar( menuBar );

        /*
         *  add a listener for CLOSE events
         */
        WindowListener wl = new WindowAdapter()
        {
            public void windowClosing( WindowEvent we )
            {
                shutDown();
            }
        };
        addWindowListener( wl );

        if ( isEpoc ) {
            /*
             *  initialize window handle in EpocUtil
             */
            EpocUtil.setWindow();

            /*
             *  guess the help file name
             */
            String dir = "Z:\\System\\Data\\";
            String[] names = { dir + "Help", dir + "Hilfe" };
            for ( int i = 0; i < names.length; ++i ) {
                if ( new File( names[ i ] ).canRead() ) {
                    helpDocument = names[ i ];
                    break;
                }
            }
        }
    }


    /**
     *  Add a toolbar
     */
    public void addToolbar( String[] buttons )
    {
        addToolbar( getTitle(), buttons );
    }


    /**
     *  Add a toolbar
     */
    public void addToolbar( String title, String[] buttons )
    {
        toolbar = new Toolbar( title, buttons );
        super.add( toolbar, BorderLayout.EAST );
    }


    /**
     *  Add user menu items - override to add your own menu items on top
     *                        or add your own menues before this one
     */
    public void addUserItems( Menu m )
    {
    }


    /**
     *  Add a menu - override to add your own menu items on the bottom
     *               or add your own menues after this one
     */
    public void addMenu( Menu m )
    {
        if ( m.getItemCount() > 0 ) {
            m.setFont( MENU_FONT );
            menuBar.add( m );
        }
    }


    /**
     *  Add a menu item
     */
    public MenuItem addMenuItem( Menu m, String label )
    {
        return addMenuItem( m, label, 0, false );
    }


    /**
     *  Add a menu item
     */
    public MenuItem addMenuItem( Menu m, String label, int key )
    {
        return addMenuItem( m, label, key, false );
    }


    /**
     *  Add a menu item
     */
    public MenuItem addMenuItem( Menu m, String label, int key, boolean shift )
    {
        MenuItem item;

        if ( key != 0 ) {
            item = new MenuItem( label, new MenuShortcut( key, shift ) );
        }
        else {
            item = new MenuItem( label );
        }
        item.addActionListener( this );
        m.add( item );
        item.setFont( MENU_FONT );

        return item;
    }


    /**
     *  set the Help file
     */
    public void setHelpDocument( String help )
    {
        try {
            helpDocument = new File( help ).getCanonicalPath();
        }
        catch ( IOException ioe ) {
            helpDocument = help;
        }
    }


    /**
     *  Adjust size and position when first shown
     */
    public void setVisible( boolean vis )
    {
        if ( vis ) {
            if ( isEpoc ) {
                setBounds( frameX, frameY, frameWidth, frameHeight );
            }
            else {
                super.setVisible( true );
                Rectangle bounds = getBounds();
                Rectangle inner = contentPanel.getBounds();

                Insets insets = new Insets( inner.y - bounds.y,
                                            inner.x - bounds.x,
                                            bounds.height - inner.height
                                            - ( inner.y - bounds.y ),
                                            inner.x - bounds.x );
                // System.out.println( insets );

                switch ( sizeTag ) {

                case TINY:
                    frameWidth = 480 + insets.left + insets.right;
                    frameHeight = 160 + insets.top + insets.bottom;
                    break;

                case NORMAL:
                    frameWidth = 640 + insets.left + insets.right;
                    frameHeight = 240 + insets.top + insets.bottom;
                    break;

                default:
                    frameWidth = 640 + insets.left + insets.right;
                    frameHeight = 480 + insets.top + insets.bottom;
                }
                setSize( frameWidth, frameHeight );
                if ( EpocUtil.isDebug() ) {
                    System.out.println( "Size: " + frameWidth
                                      + "x" + frameHeight );
                }
            }
        }
        super.setVisible( vis );
        if ( vis ) {
            contentPanel.requestFocus();
        }
    }


    /**
     *  Set window and toolbar title
     */
    public void setTitle( String title )
    {
        super.setTitle( title );
        if ( toolbar != null ) {
            toolbar.setTitle( title );
        }
    }


    /**
     *  Set a new toolbar button text
     *  @param index index of button, starting from 1
     *  @param text new label text
     */
    public void setButtonLabel( int index, String text )
    {
        if ( toolbar != null ) {
            toolbar.setButtonLabel( index, text );
        }
    }


    /**
     *  add a component to the content panel
     */
    protected void addImpl( Component component, Object constraints, int index )
    {
        if ( contentPanel != null && component != contentPanel ) {
            contentPanel.add( component, constraints, index );
        }
        else {
            super.addImpl( component, constraints, index );
        }
    }


    /**
     *  set the layout manager
     */
    public void setLayout( LayoutManager lm )
    {
        if ( contentPanel != null ) {
            contentPanel.setLayout( lm );
        }
        else {
            super.setLayout( lm );
        }
    }


    /**
     *  Action Listener.
     *  A sub class of EFrame that implements actionPerformed should call
     *  super.actionPerformed( e ), to ensure that the actions handled by
     *  EFrame are not ignored.
     */
    public void actionPerformed( ActionEvent e )
    {
        if ( e.getSource() == closeItem ) {
            shutDown();
        }
        else if ( e.getSource() == toolbarItem && toolbar != null ) {
            toolbar.setVisible( !toolbar.isVisible() );
        }
        else if ( e.getSource() == aboutItem ) {
            aboutDialog.setVisible( true );
        }
        else if ( e.getSource() == helpItem ) {
            /*
             *  This should launch a help file for the application!
             */
            help();
        }
    }


    /**
     *  Provides the about dialog
     */
    public class AboutDialog extends Dialog implements ActionListener
    {
        Button continueB = new Button( "Continue" );
        public String[] text = { "" };
        final int WIDTH = 300;
        final int HEIGHT = 100;

        AboutDialog( Frame parent )
        {
            super( parent );
            this.setLayout( new FlowLayout() );
            GridPanel panel = new GridPanel();
            Canvas canvas = new Canvas()
            {
                public void paint( Graphics g )
                {
                    paintText( g );
                }
            };
            canvas.setSize( WIDTH, HEIGHT );
            panel.add( canvas, 0, 0 );
            panel.add( continueB, 0, 1 );
            continueB.addActionListener( this );
            add( panel );
            pack();
        }

        public void actionPerformed( ActionEvent ae )
        {
            if ( ae.getSource() == continueB ) {
                this.setVisible( false );
            }
        }

        public void paintText( Graphics g )
        {
            FontMetrics fm = g.getFontMetrics();
            int lineCount = text.length;
            for ( int i = 0; i < lineCount; i++ ) {
                int x = ( WIDTH - fm.stringWidth( text[ i ] ) ) / 2;
                g.drawString( text[ i ], x,
                              HEIGHT * ( i + 1 ) / ( lineCount + 1 ) );
            }
        }
    }


    /**
     *  Simple toolbar, similar to EICON toolbar
     */
    public class Toolbar extends GridPanel implements Runnable
    {
        private TTitle title;
        private TButton buttons[];
        private TDateTime clockLabel;
        private TDateTime dateLabel;

        private final int DT_TIME = 0;
        private final int DT_DATE = 1;
        private final int DT_TOGGLE = 2;

        private final Insets INSETS = new Insets( 1, 0, 1, 0 );

        private final Dimension[] buttonDimension = {
            new Dimension( 49, 35 ),
            new Dimension( 69, 37 ),
            new Dimension( 69, 48 )
        };
        private final Dimension[] titleDimension = {
            new Dimension( 47, 18 ),
            new Dimension( 67, 23 ),
            new Dimension( 67, 48 )
        };
        private final Dimension[] labelDimension = {
            new Dimension( 47, 24 ),
            new Dimension( 67, 24 ),
            new Dimension( 67, 24 )
        };
        private final Color[] backgroundColor = {
            Color.white,
            Color.white,
            new Color( 206, 206, 206 )
        };
        private final Color[] buttonColor = {
            Color.lightGray,
            Color.lightGray,
            new Color( 165, 154, 206 )
        };
        private final int[] buttonCount = { 3, 4, 6 };


        /**
         *  Toolbar Title Class
         */
        public class TTitle extends Container
        {
            private Label label;

            /**
             *  create label and install listener
             */
            public TTitle( String text )
            {
                super();
                super.setLayout( new BorderLayout() );
                label = new Label( text, Label.CENTER );
                add( label, BorderLayout.CENTER );
                MouseListener ml = new MouseAdapter() {
                    public void mousePressed( MouseEvent e )
                    {
                        /*
                         *  display tasklist (if possible) on click
                         */
                        taskList();
                    }
                };
                addMouseListener( ml );
                label.addMouseListener( ml );
            }


            /**
             *  change the text
             */
            public void setText( String text )
            {
                label.setText( text );
            }


            /**
             *  layout hint
             */
            public Dimension getPreferredSize()
            {
                return titleDimension[ sizeTag ];
            }


            /**
             *  layout hint
             */
            public Dimension getMinimumSize()
            {
                return titleDimension[ sizeTag ];
            }
        }


        /**
         *  Toolbar Button Class
         */
        public class TButton extends Button implements ActionListener
        {
            private int index;

            /**
             *  create button and add listener
             */
            public TButton( int index, String label )
            {
                super( label );
                this.index = index;
                addActionListener( this );
            }


            /**
             *  tell application that a button is pressed
             */
            public void actionPerformed( ActionEvent e )
            {
                buttonPressed( index );
            }


            /**
             *  layout hint
             */
            public Dimension getPreferredSize()
            {
                return buttonDimension[ sizeTag ];
            }


            /**
             *  layout hint
             */
            public Dimension getMinimumSize()
            {
                return buttonDimension[ sizeTag ];
            }
        }


        /**
         *  Toolbar Date or Time Class
         */
        public class TDateTime extends Label
        {
            private DateFormat format;
            private DateFormat dateFormat;
            private DateFormat timeFormat;
            private int type;

            /**
             *  constuct label and create format strings
             */
            public TDateTime( int type )
            {
                super( "", Label.CENTER );
                this.type = type;

                switch ( type ) {

                case DT_TIME:
                    format = DateFormat.getTimeInstance( DateFormat.SHORT );
                    break;

                case DT_DATE:
                    format = new SimpleDateFormat( "EE dd" );
                    break;

                case DT_TOGGLE:
                    timeFormat = DateFormat.getTimeInstance( DateFormat.SHORT );
                    dateFormat = new SimpleDateFormat( "EE dd" );
                    format = timeFormat;

                    MouseListener ml = new MouseAdapter() {
                        public void mousePressed( MouseEvent e ) {
                            clicked();
                        }
                    };
                    addMouseListener( ml );
                }
            }


            /**
             *  toggle display
             */
            private void clicked()
            {
                if ( format == dateFormat ) {
                    format = timeFormat;
                }
                else {
                    format = dateFormat;
                }
                updateClock();
            }


            /*
             *  update the clock
             */
            public void updateClock()
            {
                Date now = new Date();
                String txt = format.format( now );
                if ( txt.length() > 6 ) {
                    int p = txt.lastIndexOf( ' ' );
                    if ( p > 0 ) {
                        txt = txt.substring( 0, p )
                            + txt.substring( p + 1 ).toLowerCase();
                    }
                }
                setText( txt );
            }


            /**
             *  layout hint
             */
            public Dimension getPreferredSize()
            {
                return labelDimension[ sizeTag ];
            }


            /**
             *  layout hint
             */
            public Dimension getMinimumSize()
            {
                return labelDimension[ sizeTag ];
            }
        }


        /**
         *  Construct Toolbar
         *  @param title caption text
         *  @param buttons array of button titles
         */
        public Toolbar( String title, String[] buttons )
        {
            setBackground( backgroundColor[ sizeTag ] );

            constraints.insets = new Insets( 0, 0, 0, 0 );
            constraints.fill = GridBagConstraints.NONE;
            constraints.anchor = GridBagConstraints.CENTER;
            constraints.gridheight = 1;
            constraints.gridwidth = GridBagConstraints.REMAINDER;

            if ( sizeTag == TINY ) {
                setFont( new Font( "Dialog", Font.BOLD, 10 ) );
            }
            else {
                setFont( new Font( "Dialog", Font.BOLD, 12 ) );
            }
            this.title = new TTitle( title );
            add( this.title, 0, 0 );

            constraints.fill = GridBagConstraints.BOTH;
            int nButtons = buttons.length;
            if ( nButtons > buttonCount[ sizeTag ] ) {
                nButtons = buttonCount[ sizeTag ];
            }
            this.buttons = new TButton[ nButtons ];
            for ( int i = 0; i < nButtons; ++i ) {
                TButton b = new TButton( i + 1, buttons[ i ] );
                this.buttons[ i ] = b;
                b.setBackground( buttonColor[ sizeTag ] );
                add( b, 0, i + 1 );
            }

            /*
             *  Spacer
             */
            Label spacer = new Label( "" );
            spacer.setBackground( backgroundColor[ sizeTag ] );
            constraints.weighty = 2;
            constraints.fill = GridBagConstraints.NONE;
            add( spacer, 0, nButtons + 1 );

            /*
             *  Running clock
             */
            if ( sizeTag == TINY ) {
                clockLabel = new TDateTime( DT_TOGGLE );
            }
            else {
                clockLabel = new TDateTime( DT_TIME );
            }
            clockLabel.setBackground( backgroundColor[ sizeTag ] );
            clockLabel.setFont( new Font( "Dialog", Font.BOLD, 16 ) );
            constraints.weighty = 0;
            constraints.fill = GridBagConstraints.NONE;
            constraints.anchor = GridBagConstraints.SOUTH;
            add( clockLabel, 0, nButtons + 2 );

            if ( sizeTag != TINY ) {
                dateLabel = new TDateTime( DT_DATE );
                dateLabel.setBackground( backgroundColor[ sizeTag ] );
                add( dateLabel, 0, nButtons + 3 );
                dateLabel.setFont( new Font( "Dialog", Font.PLAIN, 14 ) );
            }

            /*
             *  run clock
             */
            new Thread( this ).start();
        }


        /**
         *  Room for border
         */
        public Insets getInsets()
        {
            return INSETS;
        }


        /*
         *  draw a rectangle around the toolbar
         */
        public void paint( Graphics g )
        {
            super.paint( g );
            Dimension size = getSize();

            g.setColor( Color.black );
            g.drawRect( 0, 0, size.width - 1, size.height - 1 );
        }


        /*
         *  update clock every second
         */
        public void run()
        {
            while ( true ) {
                if ( isVisible() && getParent() != null
                    && getParent().isVisible() )
                {
                    updateClock();
                }
                try {
                    Thread.sleep( 1000 );
                }
                catch ( Exception e ) {}
            }
        }


        /*
         *  update clock
         */
        public void updateClock()
        {
            Date date = new Date();
            if ( clockLabel.isVisible() ) {
                clockLabel.updateClock();
            }
            if ( dateLabel != null && dateLabel.isVisible() ) {
                dateLabel.updateClock();
            }
        }


        /**
         *  set a new title
         */
        public void setTitle( String title )
        {
            this.title.setText( title );
        }


        /**
         *  set a new button text
         *  @param index index of button, starting from 1
         *  @param text new label text
         */
        public void setButtonLabel( int index, String text )
        {
            buttons[ index - 1 ].setLabel( text );
        }
    }


    /**
     *  Override shutDown() for different or additional behaviour!
     *  This implementation simply exits.
     */
    public void shutDown()
    {
        new Thread() {
            public void run() {
                try {
                    sleep( 300 );
                }
                catch ( InterruptedException ie ) {}
                System.exit( 0 );
            }
        }.start();
    }


    /**
     *  Override taskList to display a task list (how?)
     */
    public void taskList()
    {
        EpocUtil.displayTaskList();
    }


    /**
     *  Override buttonPressed for toolbar actions
     *  @param index # of the button, starting with 1
     */
    public void buttonPressed( int index )
    {
        contentPanel.requestFocus();
    }


    /**
     *  Override help for displaying a help file
     */
    public void help()
    {
        if ( helpDocument != null ) {
            try {
                EpocUtil.openDocument( helpDocument );
            }
            catch ( java.io.IOException ioe ) {
                ioe.printStackTrace();
            }
        }
    }


    /**
     *  Debug output to the console
     */
    public static void debug( Object o )
    {
        EpocUtil.debug( o );
    }

}
