/****************************************************************************

Package:            de.mvcsys.smtpauth
Class:              Sink

Task:               A dummy listener on an arbitrary port.
                    This prevents EPOC systems from crashing on modem
                    connections

Author:             Marcus von Cube

Credits:            Idea by Hans Lub (hlub@knoware.nl)

Updates:

Datum     Name  Vers.  Bemerkung
-----------------------------------------------------------------------------
21.11.03  MvC   0.6    created
21.04.05  MvC   0.7    netDial

****************************************************************************/

package de.mvcsys.smtpauth;

import de.mvcsys.util.*;
import de.mvcsys.epocutil.*;
import java.net.*;
import java.io.*;
import java.util.Vector;

/**
 *  A dummy listener on an arbitrary port.
 *  This prevents EPOC systems from crashing on modem
 *  connections.
 */
public class Sink extends Thread
{
    /**
     *  flag for activation (global for all threads)
     */
    private static boolean active;

    /**
     *  Semaphore for activation
     */
    private static Object semaphore = new Object();

    /**
     *  static output writer
     */
    private static PrintWriter sOut;

    /**
     *  instance counter
     */
    private static int nextInstance;

    /**
     *  things are different on EPOC
     */
    private static boolean noSocketTimeout = EpocUtil.isEpoc();

    /*
     *  private member variables
     */
    private String name;
    private int port;
    private PrintWriter out;
    private int instance;

    private Socket sock;

    private boolean running;
    private boolean accepting;

    /**
     *  construct with given parameters
     *  @param port port number this instance listens on
     *  @param out logging channel or <code>null</code> for System.out
     */
    public Sink( int port, PrintWriter out )
    {
        this.name = "Sink " + port;
        this.port = port;
        if ( port <= 0 ) {
            throw new IllegalArgumentException( "port: " + port );
        }
        this.out = out == null
                ? new PrintWriter( new OutputStreamWriter( System.out ) )
                : out;
        if ( sOut == null ) {
            sOut = this.out;
        }
        this.instance = nextInstance++;
    }


    /**
     *  print a message
     */
    private void log( Object o )
    {
        log( out, name, o );
    }


    /**
     *  print a message (static version)
     */
    private static void log( PrintWriter out, Object o )
    {
        log( out, "Sink", o );
    }


    /**
     *  print a message (static version)
     */
    private static void log( PrintWriter out, String name, Object o )
    {
        String line = "[" + name + "] " + o;
        if ( out == null ) {
            out = new PrintWriter( new OutputStreamWriter( System.out ) );
        }
        if ( line.endsWith( "\n" ) ) {
            out.print( line );
        }
        else {
            out.println( line );
        }
        out.flush();
    }


    /**
     *  shutdown
     */
    public void interrupt()
    {
        running = false;
        super.interrupt();

        if ( noSocketTimeout ) {
            /*
             *  on EPOC, socket timeout does not work as expected
             */
            if ( accepting ) {
                try {
                    new Socket( "localhost", port ).close();
                }
                catch ( Exception e ) {
                    log( e );
                }
            }
            else {
                try {
                    if ( sock != null ) sock.close();
                }
                catch ( Exception e ) {
                    log( e );
                }
            }
        }
    }


    /**
     *  run loop
     */
    public void run()
    {
        running = true;
        ServerSocket ss = null;

        if ( !active ) {
            log( "idle" );
            while ( !active ) {
                /*
                 *  wait for activation
                 */
                try {
                    synchronized( semaphore ) {
                        semaphore.wait();
                    }
                }
                catch ( Exception e ) {
                    log( e );
                }

                if ( !running ) {
                    log( "stopped... never active" );
                    return;
                }
            }
        }
        try {
            sleep( 300 * instance );
        }
        catch ( Exception e ) {
        }

        log( "activated" );
        try {
            /*
             *  we are listening
             */
            ss = new ServerSocket( port, 1 );
            if ( !noSocketTimeout ) {
                ss.setSoTimeout( 200 );
            }

            while ( running ) {

                sock = null;

                try {
                    /*
                     *  handle connection
                     */
                    while ( sock == null ) {
                        try {
                            accepting = true;
                            sock = ss.accept();
                            accepting = false;
                        }
                        catch ( InterruptedIOException iioe ) {}
                        catch ( SocketException se ) {}

                        if ( isInterrupted() ) {
                            throw new InterruptedException( "accept" );
                        }
                    }
                    log( "connection attempt by " + sock.getInetAddress() );

                    /*
                     *  just disconnect
                     */
                    sock.close();
                    sock = null;
                }
                catch ( InterruptedException ie ) {
                    // log( "thread interrupted" );
                }
                catch ( Throwable t ) {
                    log( "thread aborted" );
                    t.printStackTrace( out );
                    if ( running ) {
                        try {
                            sleep( 1000 );
                        }
                        catch ( Throwable th ) {}
                    }
                }
                finally {
                    /*
                     *  close sockets
                     */
                    if ( sock != null ) {
                        sock.close();
                        log( "socket closed" );
                    }
                }
            }
        }
        catch ( InterruptedIOException iie2 ) {
            log( "got interrupt on I/O (ServerSocket)" );
        }
        catch ( Throwable t2 ) {
            log( "aborted (ServerSocket)" );
            t2.printStackTrace( out );
        }
        finally {
            /*
             *  close server socket
             */
            if ( ss != null ) {
                try {
                    ss.close();
                }
                catch ( IOException ioe ) {
                    log ( ioe );
                }
            }
        }
        log( "stopped" );
    }


    /**
     *  Activate all threads
     */
    public static void activate()
    {
        if ( nextInstance == 0 ) {
            log( sOut, "nothing to activate" );
            return;
        }
        if ( active ) {
            log( sOut, "already active" );
            return;
        }

        log( sOut, "activating" );
        active = true;
        try {
            log( sOut, "My address is " + EpocUtil.netDial() );
            synchronized( semaphore ) {
                semaphore.notifyAll();
            }
        }
        catch ( Exception e ) {
            log( sOut, e );
        }
    }


    /**
     *  Is Sink configured?
     */
    public static boolean isConfigured()
    {
        return nextInstance > 0;
    }
}

