/****************************************************************************

Package:            de.mvcsys.smtpauth
Class:              SmtpAuth

Task:               SMTP proxy allowing to authenticate against server

Author:             Marcus von Cube

Credits:            Base64 borrowed from Bob Withers (bwit@pobox.com)

Updates:

Datum     Name  Vers.  Bemerkung
-----------------------------------------------------------------------------
26.04.03  MvC   0.1    created
25.05.03  MvC   0.42   MS server / yahoo tested
04.05.03  MvC   0.5    POP
05.05.03  MvC   0.51   catch more exceptions
05.03.04  MvC   0.70   check/modify mail headers
                       trace can now be configured
09.08.04  MvC   0.71   bugfix in Worker
19.04.05  MvC   0.80   EHLO with correct address, netDial
24.04.05  MvC   0.81   hostname for EHLO is now configurable

****************************************************************************/

package de.mvcsys.smtpauth;

import de.mvcsys.epocutil.*;
import de.mvcsys.util.*;
import java.util.*;
import java.io.*;

/**
 *  SMTP proxy allowing to authenticate against server
 *  Main Class
 */
public class SmtpAuth implements Runnable
{
    public static final String NAME = "SmtpAuth";
    public static final String TITLE = "SMTP Authenticating Proxy";
    public static final String VERSION = "Version 0.81, 24 Apr 2005";
    public static final String AUTHOR = "Marcus von Cube <marcus@mvcsys.de>";

    public static final int SMTP_PORT = 25;
    public static final int POP_PORT = 110;

    private boolean running;
    private Thread myThread;
    private IniFile ini;

    public PrintWriter out;
    public String iniName;  // accessed by GUI

    /**
     *  Run with System.out as console
     */
    public SmtpAuth( String[] args )
    {
        this( null, args );
    }


    /**
     *  Run with PrintWriter as console
     */
    public SmtpAuth( PrintWriter out, String[] args )
    {
        if ( out == null ) {
            out = new PrintWriter( new OutputStreamWriter( System.out ) );
        }
        this.out = out;
        log( NAME + " - " + TITLE );
        log( VERSION );
        log( "EpocUtil: " + EpocUtil.VERSION );
        log( AUTHOR );
        log( "This is free software, "
                   + "use or modify it at your own risk." );
        log( "Commercial USE is allowed but don't make a commercial "
                   + "PRODUCT from the sources!" );
        log( "" );

        iniName = args.length > 0 ? args[ 0 ] : NAME + ".ini";
        File f = new File( iniName );
        try {
            if ( !f.canRead() ) {
                /*
                 *  setup example configuration
                 */
                new File( f.getPath() ).mkdirs();
                IniFile ini = new IniFile( "example.ini" );
                ini.save( iniName );
            }
            iniName = f.getCanonicalPath();
        }
        catch ( IOException ioe ) {
            log( ioe );
            ioe.printStackTrace();
        }
    }


    /**
     *  Start the loop
     */
    public void run()
    {
        log( "Starting..." );

        running = true;
        myThread = Thread.currentThread();

        /*
         *  Read ini file
         */
        debug( "reading \"" + iniName + "\"" );
        try {
            ini = new IniFile( iniName );
        }
        catch ( IOException e ) {
            log( e );
            log( "Stopped!" );
            log( "" );
            running = false;
            return;
        }

        Enumeration sections = ini.keys();
        Vector vt = new Vector( ini.size() );
        int count = 0;
        String interf = ini.get( "Global", "interface", "localhost" );
        String hostname = ini.get( "Global", "hostname" );

        while ( sections.hasMoreElements() ) {
            /*
             *  parse sections in INI file
             */
            ++count;
            String section = (String) sections.nextElement();
            debug( "[" + section + "]" );
            if ( section.equals( "Sink" ) ) {
                /*
                 *  get state and ports for sinks (dummy listeners)
                 */
                boolean active = ini.getBoolean( section, "active", false );
                String ports = ini.get( section, "ports", "" );
                StringTokenizer t = new StringTokenizer( ports, ", " );
                while ( t.hasMoreTokens() ) {
                    try {
                        int port = Integer.parseInt( t.nextToken() );
                        /*
                         *  start a sink
                         */
                        Sink sink = new Sink( port, out );
                        sink.start();
                        vt.addElement( sink );
                    }
                    catch ( Exception e ) {
                        log( "Sink: " + e );
                    }
                }
                if ( active ) {
                    /*
                     *  let go!
                     */
                    Sink.activate();
                }
                continue;
            }

            /*
             *  handle active Server sections
             */
            if ( !section.startsWith( "Server " ) ) {
                continue;
            }
            if ( !ini.getBoolean( section, "active", true ) ) {
                continue;
            }
            String name = section.substring( 7 );
            int listenPort = ini.getInt( section, "listen-port", SMTP_PORT );
            String server = ini.get( section, "server" );
            if ( server == null ) {
                continue;
            }
            int port = ini.getInt( section, "port", SMTP_PORT );

            String popServer = null;
            int popPort = 0;
            String user = null;
            String password = null;
            boolean transparent = ini.getBoolean( section, "transparent" );

            if ( !transparent ) {
                popServer = ini.get( section, "pop-server" );
                popPort = ini.getInt( section, "pop-port", POP_PORT );
                user = ini.get( section, "user" );
                if ( user == null ) {
                    user = "-";
                }
                password = ini.get( section, "password" );
                if ( password != null ) {
                    ini.put( section, "password", password, true ); // encode!
                }
            }

            /*
             *  enable trace
             */
            OutputStream trace = null;
            if ( ini.getBoolean( "Global", "trace", false ) ) {
                String tracefile = ini.get( "Global", "trace-file" );
                if ( tracefile == null ) {
                    if ( new File( "C:\\LOGS\\Email" ).isDirectory() ) {
                        tracefile = "C:\\LOGS\\Email\\SmtpAuth.trc";
                    }
                }
                if ( tracefile != null ) {
                    try {
                        trace = new FileOutputStream( tracefile );
                        log( "Tracefile: " + tracefile );
                    }
                    catch ( Exception e ) {
                        log( e );
                    }
                }
            }

            /*
             *  create and start working thread
             */
            debug( "create Worker" );
            Worker w = new Worker( name,
                                   interf, listenPort,
                                   hostname,
                                   server, port,
                                   popServer, popPort,
                                   user, password,
                                   ini.getSection( section ),
                                   out, trace );
            debug( "start Worker" );
            w.start();
            vt.addElement( w );

            try {
                Thread.sleep( 300 );
            }
            catch ( Exception e ) {}
        }

        if ( count == 0 ) {
            /*
             *  no configuration
             */
            log( "\nNo valid setup found!" );
            log( "\nChange settings and restart!" );
            log( "" );
            running = false;
            return;
        }

        /*
         *  copy the threads into an array
         */
        Thread[] workers = new Thread[ vt.size() ];
        vt.copyInto( workers );
        vt = null;

        /*
         *  rewrite INI file with all passwords encoded
         */
        try {
            ini.save();
        }
        catch ( IOException e ) {}

        /*
         *  wait for interrupt
         */
        while ( running ) {
            try {
                Thread.sleep( 10000 );
            }
            catch( InterruptedException e ) {}
        }

        /*
         *  interrupt all threads
         */
        log( "Stopping all worker threads..." );

        for ( int i = 0; i < workers.length; ++i ) {
            workers[ i ].interrupt();
        }
        for ( int i = 0; i < workers.length; ++i ) {
            try {
                workers[ i ].join();
            }
            catch ( InterruptedException ie ) {}
        }
        log( "Engine Stopped!" );
        log( "" );
    }


    /**
     *  Shut down
     */
    public void shutDown()
    {
        if ( running ) {
            log( "Shutting down..." );
            running = false;
            myThread.interrupt();
        }
    }


    /**
     *  log
     */
    protected void log( Object o )
    {
        out.println( o );
        out.flush();
    }


    /**
     *  Debugging to the console
     */
    public static void debug( Object o )
    {
        EpocUtil.debug( o );
    }


    /**
     *  Access to ini file
     */
    public IniFile getIni()
    {
        return ini;
    }


    /**
     *  main
     */
    public static void main( String[] args )
        throws Exception
    {
        /*
         *  run the program using the console
         */
        SmtpAuth engine = new SmtpAuth( args );
        Thread t = new Thread( engine );
        t.start();

        System.out.println( "\nPress any key to exit!\n" );
        System.in.read();

        engine.running = false;
        t.interrupt();
    }
}


