/****************************************************************************

Package:            de.mvcsys.smtpauth
Class:              Worker

Task:               SMTP proxy allowing to authenticate against server
                    Worker Thread

Author:             Marcus von Cube

Credits:            Base64 borrowed from Bob Withers (bwit@pobox.com)

Updates:

Datum     Name  Vers.  Bemerkung
-----------------------------------------------------------------------------
16.05.03  MvC   0.1    created
25.05.03  MvC   0.42   MS server / yahoo tested
04.05.03  MvC   0.5    POP
05.05.03  MvC   0.51   catch more exceptions
05.03.04  MvC   0.70   check/modify mail headers
09.08.04  MvC   0.71   bugfix (NullPointerExc if no AUTH and no headers)
                       header detection modified
19.04.05  MvC   0.80   EHLO with correct address
24.04.05  MvC   0.81   hostname for EHLO is now configurable

****************************************************************************/

package de.mvcsys.smtpauth;

import de.mvcsys.util.*;
import de.mvcsys.epocutil.*;
import java.net.*;
import java.io.*;
import java.util.*;

/**
 *  SMTP proxy allowing to authenticate against server
 *  worker thread
 */
public class Worker extends Thread
{
    /**
     *  default port
     */
    public static final int SMTP_PORT = 25;

    /**
     *  default port for POP before SMTP
     */
    public static final int POP_PORT = 110;

    /*
     *  private member variables
     */
    private String name;
    private String interf;
    private int listenPort;
    private String hostname;
    private String server;
    private int port;
    private String popServer;
    private int popPort;
    private String user;
    private String password;
    private Dictionary config;
    private Hashtable headers;
    private PrintWriter out;
    private OutputStream trace;
    private boolean isEpoc = EpocUtil.isEpoc();
    private boolean noSocketTimeout = isEpoc;

    private Socket localSock;
    private Socket remoteSock;

    private String localAddress;

    private InputStream inboundInput;
    private OutputStream outboundOutput;

    private InputStream outboundInput;
    private OutputStream inboundOutput;

    private Thread inThread;
    private Thread outThread;

    private boolean running;
    private boolean accepting;
    private boolean connected;
    private boolean transparent;
    private boolean waitAuth;
    private boolean doNotDisturb;
    private boolean inEnvelope;
    private boolean inHead;

    /**
     *  construct with given parameters
     *  @param name name of this instance
     *  @param interf IP this instance listens on or "all"
     *  @param listenPort port number this instance listens on
     *  @param server name or IP of SMTP server
     *  @param port port number the server is listening on
     *  @param popServer name or IP of otional POP server (POP before SMTP)
     *  @param popPort port number for POP server (POP before SMTP)
     *  @param user authenticating user name
     *  @param password authenticating password
     *  @param out logging channel or <code>null</code> for System.out
     *  @param out trace file
     */
    public Worker( String name,
                   String interf, int listenPort,
                   String hostname,
                   String server, int port,
                   String popServer, int popPort,
                   String user, String password,
                   Dictionary config,
                   PrintWriter out,
                   OutputStream trace )
    {
        this.name = name;
        this.interf = interf == null ? "localhost" : interf;
        this.listenPort = listenPort <= 0 ? SMTP_PORT : listenPort;
        this.hostname = hostname;
        this.server = server;
        this.port = port <= 0 ? SMTP_PORT : port;
        this.popServer = popServer;
        this.popPort = popPort <= 0 ? POP_PORT : popPort;
        this.user = user;
        this.password = password;
        this.out = out == null
                ? new PrintWriter( new OutputStreamWriter( System.out ) )
                : out;
        this.trace = trace;

        /*
         *  get headers from config
         */
        Enumeration keys = config.keys();
        while( keys.hasMoreElements() ) {
            String key = (String) keys.nextElement();
            if ( !key.startsWith( "header." ) ) {
                continue;
            }
            if ( headers == null ) {
                headers = new Hashtable();
            }
            String val = (String) config.get( key );
            key = key.substring( 7 ).toLowerCase();
            headers.put( key, val );
        }

        log(     "listen port: " + this.listenPort );
        log(     "server:      " + server + ":" + this.port );
        if ( popServer != null ) {
            log( "POP server:  " + popServer + ":" + this.popPort );
        }
        log(     "user name:   " + user );
    }


    /**
     *  print a message
     */
    private void log( Object o )
    {
        String line = "[" + name + "] " + o;
        if ( line.endsWith( "\n" ) ) {
            out.print( line );
        }
        else {
            out.println( line );
        }
        out.flush();
        debug( line );
    }


    /**
     *  shutdown
     */
    public void interrupt()
    {
        running = false;
        super.interrupt();

        if ( noSocketTimeout ) {
            /*
             *  on EPOC, socket timeout does not work as expected
             */
            if ( accepting ) {
                try {
                    new Socket( "localhost", listenPort ).close();
                }
                catch ( Exception e ) {
                    log( e );
                }
            }
            else {
                try {
                    if ( localSock  != null ) localSock.close();
                    if ( remoteSock != null ) remoteSock.close();
                }
                catch ( Exception e ) {
                    log( e );
                }
            }
        }
    }


    /**
     *  run loop
     */
    public void run()
    {
        ServerSocket ss = null;

        log( "started" );
        running = true;

        try {
            /*
             *  we are listening
             */
            debug( "creating ServerSocket" );
            if ( interf.equals( "all" ) || interf.equals( "0.0.0.0" )
              || interf.equals( "*" ) )
            {
                ss = new ServerSocket( listenPort, 1 );
            }
            else {
                ss = new ServerSocket( listenPort, 1,
                                       InetAddress.getByName( interf ) );
            }
            log( "listen on " + ss.getInetAddress() + ":" + ss.getLocalPort() );
            if ( !noSocketTimeout ) {
                ss.setSoTimeout( 200 );
            }

            while ( running ) {

                inThread = null;
                localSock = null;
                remoteSock = null;

                try {
                    /*
                     *  handle connection
                     */
                    log( "waiting for connection" );
                    while ( localSock == null ) {
                        try {
                            accepting = true;
                            localSock = ss.accept();
                            accepting = false;
                        }
                        catch ( InterruptedIOException iioe ) {}
                        catch ( SocketException se ) {}

                        if ( isInterrupted() ) {
                            throw new InterruptedException( "accept" );
                        }
                    }
                    log( "connection accepted" );
                    if ( !noSocketTimeout ) {
                        localSock.setSoTimeout( 200 );
                    }

                    if ( popServer != null ) {
                        /*
                         *  do a quick logon to the POP server
                         */
                        logonToPopServer();
                    }

                    /*
                     *  Dial connection
                     */
                    log( "connecting" );
                    InetAddress local = EpocUtil.netDial();
                    debug( "dial complete, local = " + local );

                    /*
                     *  connect to server
                     */
                    debug( "creating Socket" );
                    remoteSock = new Socket( server, port );
                    log( "connected to " + server + ":" + port );
                    if ( !noSocketTimeout ) {
                        remoteSock.setSoTimeout( 200 );
                    }
                    connected = true;

                    /*
                     *  Who are we?
                     */
                    if ( !isEpoc ) {
                        local = remoteSock.getLocalAddress();
                    }
                    log( "local address is " + local );
                    localAddress = local.getHostName();
                    if ( localAddress == null || localAddress.length() == 0 ) {
                        localAddress = local.getHostAddress();
                    }
                    if ( isInterrupted() ) {
                        throw new InterruptedException( "connect" );
                    }

                    /*
                     *  get All Streams
                     */
                    inboundInput = new BufferedInputStream(
                            localSock.getInputStream() );
                    outboundOutput = new BufferedOutputStream(
                            remoteSock.getOutputStream() );

                    outboundInput = new BufferedInputStream(
                            remoteSock.getInputStream() );
                    inboundOutput = new BufferedOutputStream(
                            localSock.getOutputStream() );

                    /*
                     *  create and start thread for inbound data
                     */
                    inThread = new Thread() {
                        public void run()
                        {
                            try {
                                proxy( false );
                                log( "inbound thread ended" );
                            }
                            catch ( InterruptedException ie ) {
                                log( "inbound thread interrupted" );
                            }
                            catch ( Throwable t ) {
                                log( "inbound thread aborted" );
                                t.printStackTrace( out );
                                if ( running ) {
                                    try {
                                        sleep( 1000 );
                                    }
                                    catch ( Throwable th ) {}
                                }
                            }
                            finally {
                                inThread = null;
                                if ( outThread != null ) {
                                    try {
                                        log( "interrupting outbound thread" );
                                        outThread.interrupt();
                                    }
                                    catch ( Exception ex ) {}
                                }
                            }
                        }
                    };
                    inThread.setDaemon( true );
                    inThread.start();

                    /*
                     *  create and start thread for outbound data
                     */
                    outThread = new Thread() {
                        public void run()
                        {
                            try {
                                proxy( true );
                                log( "outbound thread ended" );
                            }
                            catch ( InterruptedException ie ) {
                                log( "outbound thread interrupted" );
                            }
                            catch ( Throwable t ) {
                                log( "outbound thread aborted" );
                                t.printStackTrace( out );
                                if ( running ) {
                                    try {
                                        sleep( 1000 );
                                    }
                                    catch ( Throwable th ) {}
                                }
                            }
                            finally {
                                outThread = null;
                                if ( inThread != null ) {
                                    try {
                                        log( "interrupting inbound thread" );
                                        inThread.interrupt();
                                    }
                                    catch ( Exception ex ) {}
                                }
                            }
                        }
                    };
                    outThread.setDaemon( true );
                    outThread.start();
                    outThread.join();
                }
                catch ( InterruptedException ie ) {
                    // log( "outbound thread interrupted" );
                }
                catch ( Throwable t ) {
                    log( "outbound thread aborted" );
                    t.printStackTrace( out );
                    if ( running ) {
                        try {
                            sleep( 1000 );
                        }
                        catch ( Throwable th ) {}
                    }
                }
                finally {
                    /*
                     *  interrupt threads
                     */
                    connected = false;
                    if ( inThread != null ) {
                        inThread.interrupt();
                    }
                    if ( outThread != null ) {
                        outThread.interrupt();
                    }

                    /*
                     *  close sockets
                     */
                    if ( localSock != null ) {
                        localSock.close();
                        log( "inbound socket closed" );
                    }
                    if ( remoteSock != null ) {
                        remoteSock.close();
                        log( "outbound socket closed" );
                    }
                    /*
                     *  close trace
                     */
                    if ( trace != null ) {
                        trace.close();
                        trace = null;
                    }
                }
            }
        }
        catch ( InterruptedIOException iie2 ) {
            log( "got interrupt on I/O (ServerSocket)" );
        }
        catch ( Throwable t2 ) {
            log( "aborted (ServerSocket)" );
            t2.printStackTrace( out );
        }
        finally {
            /*
             *  close server socket
             */
            if ( ss != null ) {
                try {
                    ss.close();
                }
                catch ( IOException ioe ) {
                    log ( ioe );
                }
            }
        }
        log( "stopped" );
    }


    /**
     *  proxy implementation for a single connection
     *  routine is called by a separate thread for each direction
     */
    private void proxy( boolean outbound )
        throws IOException, InterruptedException
    {
        InputStream is = outbound ? inboundInput : outboundInput;
        OutputStream os = outbound ? outboundOutput : inboundOutput;
        byte[] buffer = new byte[ 256 ];
        int count = 0;
        boolean ok = true;
        int total = 0;
        boolean auth = false;
        boolean deleting = false;

        if ( outbound ) {
            /*
             *  transparent mode if user is null, analyse mode else
             */
            transparent = user == null;
            waitAuth = !transparent && !user.equals( "-" );
            if ( !waitAuth && headers == null ) {
                /*
                 *  nothing to check or modify
                 */
                transparent = true;
            }
            if ( transparent ) {
                log( "transparent mode" );
            }
            else {
                inEnvelope = true;
            }
            connected = true;
        }

        while ( running && connected && ok && !interrupted() ) {
            /*
             *  read and check for new line or buffer full
             */
            int c = -2;
            while ( c == -2 ) {
                try {
                    c = is.read();
                }
                catch ( InterruptedIOException ioe ) {
                    if ( interrupted() ) {
                        throw new InterruptedException( "read" );
                    }
                }
                catch ( IOException ioe ) {
                    if ( interrupted() ) {
                        throw new InterruptedException( "read" );
                    }
                    else {
                        if ( outbound ) {
                            log( "client read: " + ioe );
                        }
                        else {
                            log( "server read: " + ioe );
                        }
                        break;
                    }
                }
            }
            if ( c < 0 ) {
                /*
                 *  eof
                 */
                ok = false;
                if ( outbound ) {
                    log( "client closed connection, "
                         + total + " bytes sent" );
                }
                else {
                    log( "server closed connection, "
                         + total + " bytes received" );
                }
                continue;
            }
            buffer[ count++ ] = (byte) c;

            while ( running && outbound && doNotDisturb ) {
                /*
                 *  be patient
                 */
                try {
                    sleep( 200 );
                }
                catch ( InterruptedException ie ) {}
            }
            ++total;

            /*
             *  check for line end or full buffer
             *  just forward if in transparent mode
             */
            if ( !transparent && c == 10
                 || count == buffer.length
                 || transparent && is.available() == 0 )
            {
                /*
                 *  end of line in analyse mode
                 *  or buffer full
                 *  or no more data in transparent mode
                 */
                boolean dontSend = false;

                if ( !transparent ) {
                    /*
                     *  get the line as string to analyse it
                     */
                    String line = new String( buffer, 0, count );
                    if ( outbound ) {
                        /*
                         *  to server
                         */
                        if ( waitAuth && line.startsWith( "HELO" ) ) {
                            /*
                             *  replace HELO with EHLO
                             */
                            if ( false ) {
                                buffer[ 0 ] = 'E';
                                buffer[ 1 ] = 'H';
                            }
                            else {
                                dontSend = true;
                                if ( hostname == null ) {
                                    sendLine( "EHLO " + localAddress );
                                }
                                else {
                                    sendLine( "EHLO " + hostname );
                                }
                            }
                        }
                        else if ( inHead ) {
                            /*
                             *  we might want to delete/replace header lines
                             */
                            if ( buffer[ 0 ] == 9 || buffer[ 0 ] == ' ' ) {
                                /*
                                 *  line continuation
                                 */
                                dontSend = deleting;
                            }
                            else {
                                /*
                                 *  check for known header
                                 */
                                deleting = false;
                                int p = line.indexOf( ":" );
                                if ( p > 0 ) {
                                    String key = line.substring( 0, p )
                                                    .toLowerCase();
                                    String val = (String) headers.get( key );
                                    if ( val != null ) {
                                        /*
                                         *  skip this header
                                         */
                                        deleting = true;
                                        dontSend = true;
                                        log( "header \"" + line.trim()
                                           + "\" skipped" );
                                    }
                                }
                                /*
                                 *  check for header end (empty line)
                                 */
                                if ( line.trim().length() == 0 ) {
                                    /*
                                     * end of header block
                                     */
                                    inHead = false;
                                    if ( headers != null ) {
                                        /*
                                         * add headers
                                         */
                                        Enumeration e = headers.keys();
                                        while ( e.hasMoreElements() ) {
                                            String key = (String) e.nextElement();
                                            String s = (String) headers.get( key );
                                            if ( !s.equals( "-" ) ) {
                                                s = key.substring( 0, 1 )
                                                       .toUpperCase()
                                                  + key.substring( 1 )
                                                  + ": "
                                                  + s;
                                                sendLine( s );
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else {
                            /*
                             *  inside message, wait for "."
                             */
                            if ( buffer[ 0 ] == '.'
                              && line.trim().equals( "." ) )
                            {
                                /*
                                 *  end of message text
                                 */
                                inEnvelope = true;
                            }
                        }
                    }
                    else {
                        /*
                         *  from server
                         */
                        if ( line.startsWith( "250-AUTH" )
                             && line.indexOf( "LOGIN" ) >= 0 )
                        {
                            /*
                             *  authentication
                             */
                            auth = true;
                            waitAuth = false;
                        }
                        else if ( auth && line.startsWith( "250 " ) ) {
                            /*
                             *  start authentication
                             */
                            doNotDisturb = true;
                            sendLine( "AUTH LOGIN" );
                        }
                        else if ( line.startsWith( "334 " ) ) {
                            /*
                             *  check for Username/Password prompt
                             */
                            String prompt = decodeLine( line );
                            log( prompt );
                            if ( prompt.indexOf( "name" ) >= 0 ) {
                                /*
                                 *  Username
                                 */
                                dontSend = true;
                                sendLine( Base64.encode( user ) );
                                log( user );
                            }
                            else if ( prompt.indexOf( "word" ) >= 0 ) {
                                /*
                                 *  Password
                                 */
                                dontSend = true;
                                sendLine( Base64.encode( password ) );
                                log( "<password>" );
                            }
                        }
                        else if ( line.startsWith( "235 " ) ) {
                            /*
                             *  done
                             */
                            dontSend = true;
                            log( line );
                            if ( headers == null ) {
                                transparent = true;
                            }
                            auth = false;
                            doNotDisturb = false;
                        }
                        else if ( line.startsWith( "354 " ) ) {
                            /*
                             *  prompt to send mail text
                             */
                            if ( waitAuth ) {
                                /*
                                 *  no authentication requested
                                 */
                                log( "no authentication requested" );
                                waitAuth = false;

                                if ( headers == null ) {
                                    /*
                                     *  no header checking,
                                     *  switch to transparent mode
                                     */
                                    transparent = true;
                                }
                            }
                            inEnvelope = false;
                            inHead = !transparent;
                        }
                        else if ( line.startsWith( "4" ) ) {
                            /*
                             *  warning
                             */
                            log( line );
                        }
                        else if ( line.startsWith( "5" ) ) {
                            /*
                             *  failure
                             */
                            log( line );
                            transparent = true;
                            doNotDisturb = false;
                        }
                    }
                }
                if ( !dontSend ) {
                    /*
                     *  forward data
                     */
                    os.write( buffer, 0, count );
                    os.flush();
                    if ( trace != null ) {
                        trace.write( buffer, 0, count );
                    }
                }
                count = 0;
            }
        }
        if ( interrupted() ) {
            throw new InterruptedException();
        }
    }


    /*
     *  send a line to the remote Host
     */
    private void sendLine( String line )
            throws IOException
    {
        byte[] b = ( line + "\r\n" ).getBytes();
        outboundOutput.write( b );
        outboundOutput.flush();

        if ( trace != null ) {
            if ( line.startsWith( "PASS " + password ) ) {
                b = "PASS <password>\r\n".getBytes();
            }
            trace.write( b );
            trace.flush();
        }

        // log( Thread.currentThread() + ": " + new String( b ) );
    }


    /*
     *  decode a line into a String (strip number)
     */
    private static String decodeLine( String line )
    {
        int poslf = line.indexOf( "\n" );
        int poscr = line.indexOf( "\r" );
        int end;
        int beg;

        if ( poscr == poslf - 1 ) {
            end = poscr;
        }
        else if ( poslf >= 0 ) {
            end = poslf;
        }
        else {
            end = line.length();
        }
        if ( line.startsWith( "334 " ) ) {
            beg = 4;
        }
        else {
            beg = 0;
        }

        return Base64.decode( line.substring( beg, end ) );
    }


    /*
     *  logon to POP server for POP before SMTP
     */
    private void logonToPopServer()
            throws IOException, InterruptedException
    {
        /*
         *  connect to server
         */
        remoteSock = new Socket( popServer, popPort );
        log( "connected to " + popServer + ":" + popPort );
        if ( !noSocketTimeout ) {
            remoteSock.setSoTimeout( 200 );
        }
        if ( isInterrupted() ) {
            throw new InterruptedException( "POP connect" );
        }
        connected = true;

        /*
         *  get streams
         */
        outboundOutput = new BufferedOutputStream(
                remoteSock.getOutputStream() );
        outboundInput = new BufferedInputStream(
                remoteSock.getInputStream() );

        /*
         *  loop
         */
        byte[] buffer = new byte[ 256 ];
        boolean ok = true;
        int state = 0;
        int count = 0;

        while ( running && ok && !interrupted() ) {
            /*
             *  read and check for new line or buffer full
             */
            int c = -2;
            while ( c == -2 ) {
                try {
                    c = outboundInput.read();
                }
                catch ( InterruptedIOException ioe ) {
                    if ( interrupted() ) {
                        throw new InterruptedException( "POP read" );
                    }
                }
                catch ( IOException ioe ) {
                    if ( interrupted() ) {
                        throw new InterruptedException( "POP read" );
                    }
                    else {
                        log( "POP read: " + ioe );
                        break;
                    }
                }
            }
            if ( c < 0 ) {
                /*
                 *  eof
                 */
                ok = false;
                log ( "POP connection closed by server" );
                continue;
            }
            buffer[ count++ ] = (byte) c;
            if ( trace != null ) trace.write( c );

            /*
             *  check for line end or full buffer
             */
            if ( c == 10 || count == buffer.length ) {
                /*
                 *  get the line as string to analyse it
                 */
                String line = new String( buffer, 0, count );

                if ( line.startsWith( "+OK" ) ) {
                    /*
                     *  send command
                     */
                    switch ( state++ ) {

                    case 0:
                        sendLine( "USER " + user );
                        log( "USER " + user );
                        break;

                    case 1:
                        sendLine( "PASS " + password );
                        log( "PASS <password>" );
                        break;

                    case 2:
                        if ( line.length() > 6 ) {
                            log( line );
                        }
                        sendLine( "STAT" );
                        log( "STAT" );
                        break;

                    case 3:
                        log( line );
                        sendLine( "QUIT " + password );
                        break;

                    default:
                        ok = false;
                    }
                }
                else {
                    /*
                     *  error from server
                     */
                    log( line );
                    ok = false;
                }
                count = 0;
            }
        }
        if ( interrupted() ) {
            throw new InterruptedException();
        }

        /*
         *  disconnect
         */
        try {
            connected = false;
            remoteSock.close();
            outboundOutput = null;
            outboundInput = null;
        }
        catch ( IOException ioe ) {
            log( "POP disconnect: " + ioe );
        }
    }


    /**
     *  Debugging to the console
     */
    public static void debug( Object o )
    {
        EpocUtil.debug( o );
    }

}
